use Parallel::ForkManager;
use Getopt::Long;
use List::Util 'shuffle';
$|=1;
$PACKEIS_version="1.0";
$PACKEIS_date="2019-01-18";

$logo=
"                                                                              
                             iBMq                                              
                            iBMBBi                                             
        YGBBBMBGoi         iBMuXBG                                        i    
      MBBBNPXqZBMBBi      iBMI LBB                                      VBBB   
     iSXr        SBM      BBO  JBM          iJXFUv        PB         iMBBMvi   
                 iMB     BBB   rMB        uBBBBBBBB      vBBv      vMBBN       
          BML    FBB    BBB    iBB       EBBr             MBO    YBBBr         
         vBB uBBBBBr   BBB  BBBBMBr     iBBi              IBM  rBBBBBi         
         OBM  Yrvi    BBB   riiiBBM     BBG               vBGiNBBY FBB         
        YBMi         BBB        LBM    rMB       iYMBv    rBBBBP    BBG        
       vBBS         BBB          i     iBBL iiuGBBBBOL    vMBBi      BBBi      
       rEO         iMBi                 vBBBMBBBNv         ir         MBM      
                                                                       v       
                                                                                                   
           rIZVIujYLuIj                               iivrii                   
          iBBBBBBBMBBBM                NGL        JBBBBBBBBB                   
          rBB                         VMBi       BBGr                          
          BBv                        iBBu       SBBi                           
         BBBLiirvJ                   BBB        iBBMBBBMBMBBBBBBBX             
        BBGMBBBBBBv                 OBB            irvuIuUuLri EMB             
       GBM                         MBBi                        MBX             
      OBG                         BBB                       iGBBV              
     BBMr iiiiiiiiiiirvrvjJ      GBB           iiiirvLuoZBBMBBX                
      MBBBBBBBBBBBBBBBBBBBBi                 BBBMBBBBBBBBGji



VERSION: .......... $PACKEIS_version
LAST MODIFIED: .... $PACKEIS_date
";

$usage=
"
===============================================================================
                               +++ USAGE +++
			      
perl PACKEIS.pl -i [input] (-option [value])

Value for -i can be a FASTA file or a plain sequence. Use -i multiple times for
multiple input files or sequences.

Contact:
David Rosenkranz
Institute of Organismic and Molecular Evolutionary Biology
Dept. Anthropology, small RNA group
Johannes Gutenberg University Mainz
email: rosenkranz\@uni-mainz.de

You can find the latest PACKEIS version at:
http://sourceforge.net/projects/packeis/files
http://www.smallRNAgroup-mainz.de/software.html

===============================================================================
                              +++ OPTIONS +++

 -input OR -i [file or string]   Name of input file(s) in FASTA format.
                                 Alternatively type sequence as plain text.
 -model OR -m [0,1,2]            Model to be used for reconstruction of
                                 artificial coding sequences. 0=free,
                                 1=strict, 2=shuffle. Default=2.
 -codon_usage OR -c [file]       File that contains user-defined codon usage.
                                 Only relevant (but optionally) in combination
                                 with -m 1.
 -mt                             Use mitochondrial genetic code. Default=off.
 -altseqs OR -a [int]            Number of alternative coding sequences to be
                                 created per input sequence. Default=100.
 -threshold OR -l [int]          Minimum length of a input sequence [nt] to
                                 run RNAplfold (local folding). Otherwise use
                                 RNAfold (global folding). Default=100.
 -keep_aCDS OR -k                Save alternative coding sequences for each
                                 input sequence. Default=off.
 -threads OR -t [int]            Number of parallel threads. Default=1.
 -silent OR -s                   Less output to STDOUT.
 -help OR -h                     Prints this screen.


";

# defaults
$model=2;
$alt_seqs=100;
$threads=1;
$threshold_length=100;

# get options
GetOptions
	(
	"i|input=s"=>\@CDS, # either plain sequence or a file that contains sequence(s) in FASTA format
	"m|model=i"=>\$model, # 0 = random codon usage | 1= codon probability according to codon usage in input fasta file | 2 = codon shuffle
	"mt"=>\$MT, # use mitochondrial genetic code instead of standard genetic code
	"a|altseqs=i"=>\$alt_seqs, # number of artificial counterparts to be created
	"c|codon_usage=s"=>\$cu_table, # name of the file that contains alternative codon frequencies
	"l|threshold=i"=>\$threshold_length, # minimum length of a input sequence to run RNAplfold (local folding). Otherwise use RNAfold (global folding)
	"t|threads=i"=>\$threads, # number of parallel threads
	"k|keep_aCDS"=>\$keep_aORF, # save alternative CDS sequences in output files
	"s|silent"=>\$silent, # do not output status information to STDOUT
	"h|help"=>\$print_help, # print help
	)||die print"\nError in command line arguments.\n\n";

print$logo;
if($print_help)
	{
	print$usage;
	exit;
	}

if($model<0||$model>2){die print"\nValue for option -m (-model) must be 0, 1 or 2.\n\n";}
if($alt_seqs<100){print"\n\nWARNING! Number of alternative CDS sequences is low (=$alt_seqs)!\nContinue nevertheless in 10 seconds...\n\n";sleep(10);}

if($MT)
	{# mtDNA genetic code
	a2c_MT();
	c2a_MT();
	$genetic_code="Mitochondrial";
	}
else
	{# standard genetic code
	a2c();
	c2a();
	$genetic_code="Standard";
	}

# open LOG file and create output folder
$localtime=localtime;
$localtime=~s/ /-/g;
$localtime=~s/:/h/;
$localtime=~s/:/m/;
$localtime=~s/-\d{4}$/s$&/;
$results_folder="PACKEIS.$localtime";
if(!-d$results_folder)
	{
	mkdir("$results_folder")||die print"\n\nERROR! Cannot create output folder $results_folder.\n$!\n\n";
	}
mkdir("$results_folder/raw");
$localtime=localtime;
open(LOG,">$results_folder/PACKEIS.log")||die print"\n\nERROR! Cannot creat LOG file.\n$!\n\n";
print LOG"Time: $localtime
Model: $model
Genetic code: $genetic_code
Alternative sequences: $alt_seqs
Threads: $threads
Min. length of a input seq. to run RNAplfold instead of RNAfold: ${threshold_length}nt\n\n";

# load input sequences
%CDS=();
%species2ID=();
%transcript2ID=();
%ID2species=();
%ID2transcript=();
%FASTA_headers=();
$species_ID=0;
$plain_sequence_id=0;
$input_seq_count=0;
if(!$silent){print"\nLoad input sequences...";}
foreach$file(@CDS)
	{
	if(-e$file)  # input is a file
		{
		open(FILE,"$file")||die print"\n\nERROR! Cannot open input file $file.\n$!\n\n";
		$spec=$file;
		$spec=~s/\.[^.]+$//; # this is actually the file's basename
		$species_ID++;
		$transcript_ID=0;
		
		# replace species and transcript name by numeric ID (RNAfold and RNALfold have an output file name length restriction)
		$species2ID{$spec}=$species_ID;
		$ID2species{$species_ID}=$spec;
		print LOG"SPEC: $spec -> $species_ID\n";
		while(<FILE>)
			{
			$_=~s/\s*$//;
			if($_=~s/^>//)
				{
				$input_seq_count++;
				
				# check if FASTA header occurs more than once
				$FASTA_headers{$spec}{$_}++;
				if($FASTA_headers{$spec}{$_}>1)
					{
					print"\n\nERROR! FASTA header $_ in $file occurs more than once.\n\n";exit;
					}
				$transcript=$_;
				$transcript_ID++;
				print LOG"TR: $transcript -> $transcript_ID\n";
				
				# replace species and transcript name by numeric ID (RNAfold and RNALfold have an output file name length restriction)
				$ID2transcript{$species_ID}{$transcript_ID}=$transcript;
				$transcript2ID{$species_ID}{$transcript}=$ID2transcript;
				}
			else
				{
				$_=~s/T/U/ig;
				$_=~tr/YRWSKMDVHBNyrwskmdvhbn/CAUGUCGCAUACAUGUCGCAUA/; # replace ambiguities
				$CDS{$species_ID}{$transcript_ID}.=uc$_; # make sure to use upper case letters only
				}
			}
		close FILE;
		}
	elsif($file=~/^[ATGCUatgcu]+$/) # input is a plain sequence
		{
		$input_seq_count++;
		$species_ID++;
		$transcript_ID=1;
		$plain_sequence_id++;
		$spec="input_seq_$plain_sequence_id";
		$transcript="na";
		$species2ID{$spec}=$species_ID;
		$ID2species{0}=$spec;
		print LOG"SPEC: $spec -> $species_ID\n";
		$seq=$file;
		$seq=~s/T/U/ig;
		$seq=~tr/YRWSKMDVHBNyrwskmdvhbn/CAUGUCGCAUACAUGUCGCAUA/; # replace ambiguities
		$CDS{$species_ID}{$transcript_ID}.=uc$seq; # make sure to use upper case letters only
		$ID2transcript{$species_ID}{1}=$transcript;
		$transcript2ID{$species_ID}{$transcript}=1;
		}
	else
		{
		print"\n\nERROR! $file is not a file nor a valid coding sequence.\n\n";exit;
		}
	}
if($input_seq_count==0){print" done.\n\nERROR! No input sequences loaded.\n";exit;}
if(!$silent){print" done. $input_seq_count sequences loaded.";}

$pm=Parallel::ForkManager->new($threads);
foreach$species_ID(sort{$a<=>$b}keys%CDS)
	{
	if($model==1) 
		{
		# load codon frequencies from file
		if($cu_table)
			{
			open(CU,$cu_table)||die print LOG"Cannot open codon usage file $cu_table. $!\n";
			%codon_frequency=();
			while(<CU>)
				{
				$_=~s/\s*$//;
				@cu=split("->",$_);
				@freq=split(":",$cu[1]);
				$codon_frequency{$freq[0]}=$freq[1];
				}
			close CU;
			}
		# check codon frequencies in input sequences and create codon usage file
		else
			{
			open(CU,">$results_folder/codon_usage.$ID2species{$species_ID}.table")||print"\n\nERROR! Cannot create codon usage file.\n$!\n\n";
			print"\nCheck codon frequencies in $ID2species{$species_ID}.";
			%codon_counts=();
			%aa_counts=();
			foreach$transcript_ID(sort{$a<=>$b}keys%{$CDS{$species_ID}})
				{
				$codon_start=-3;
				while(1)
					{
					$codon_start+=3;
					$codon_counts{substr($CDS{$species_ID}{$transcript_ID},$codon_start,3)}++;
					$aa_counts{$c2a{substr($CDS{$species_ID}{$transcript_ID},$codon_start,3)}}++;
					last if(!$c2a{substr($CDS{$species_ID}{$transcript_ID},$codon_start,3)}); # stop with stop codon
					last if($codon_start>=length$CDS{$species_ID}{$transcript_ID}); # stop at end of transcript
					}
				}
			%codon_frequency=();
			foreach$aa(sort{$a cmp $b}keys%a2c)
				{
				foreach$codon(@{$a2c{$aa}})
					{
					$codon_frequency{$codon}=$codon_counts{$codon}/$aa_counts{$aa};
					$freq=$codon_counts{$codon}/$aa_counts{$aa};
					if(!$silent){print"\n$aa->$codon:$freq";}
					print CU"$aa->$codon:$freq\n";
					}
				}
			undef%codon_counts;
			undef%aa_counts;
			close CU;
			}
		}
	
	foreach$transcript_ID(sort{$a<=>$b}keys%{$CDS{$species_ID}})
		{
		# check if first codon is START CODON
		next if($CDS{$species_ID}{$transcript_ID}!~/^AUG/&&!$MT); 
		next if($CDS{$species_ID}{$transcript_ID}!~/^AUG/&&$CDS{$species_ID}{$transcript_ID}!~/^AUA/&&$MT);
		$pm->start and next;
		
		# construct protein sequence from CDS
		if(!$silent){print"\n\n$ID2species{$species_ID} -> $ID2transcript{$species_ID}{$transcript_ID} (ID:$transcript_ID)";}
		$nucleotide_sequence=$CDS{$species_ID}{$transcript_ID};
		$protein_sequence="";
		$codon_start=-3;
		%codon_dump=(); # store codons for codon shuffle mode
		while(1)
			{
			$codon_start+=3;
			$protein_sequence.=$c2a{substr($nucleotide_sequence,$codon_start,3)};
			
			if($model==2&&$c2a{substr($nucleotide_sequence,$codon_start,3)}) # store all codons from this transcript for codon shuffle mode
				{
				$codon_dump{$c2a{substr($nucleotide_sequence,$codon_start,3)}}.="#".substr($nucleotide_sequence,$codon_start,3);
				}
			last if(!$c2a{substr($nucleotide_sequence,$codon_start,3)}); # stop with stop codon
			last if($codon_start>=length$nucleotide_sequence); # stop at end of transcript
			}
		
		# adjust length of input nucleotide sequence for folding to CDS length -> replicate CDS will have same length compared to original CDS
		$nucleotide_sequence=substr($nucleotide_sequence,0,3*length$protein_sequence);
		
		if(!$silent){print"\nCDS:\n$nucleotide_sequence\nPROTEIN:\n$protein_sequence\n\n";}
		
		# calculate paired bases for original CDS and alternative CDS sequences
		open(OUT,">$results_folder/raw/$ID2species{$species_ID}.$transcript_ID.results")||print LOG"Failed to create file $results_folder/raw/$ID2species{$species_ID}.$transcript_ID.results. $!\n";
		open(ORIG_CDS,">$species_ID.$transcript_ID.cds")||print LOG"Failed to create temporary input file $species_ID.$transcript_ID.cds. $!\n";
		print ORIG_CDS">$species_ID.$transcript_ID.cds\n$nucleotide_sequence";
		close ORIG_CDS;
		
		# exit this fork if file opening failed
		if(!-e"$results_folder/raw/$ID2species{$species_ID}.$transcript_ID.results"||!-e"$species_ID.$transcript_ID.cds")
			{
			$pm->finish;
			}
		
		# switch to RNAfold for short sequences
		$folding_tool=2;
		if(length$nucleotide_sequence<$threshold_length)
			{
			$folding_tool=1;
			}
		
		if($folding_tool==1) # use RNAfold
			{
			system("RNAfold < $species_ID.$transcript_ID.cds > $species_ID.$transcript_ID.RNAfold");
			unlink"$species_ID.$transcript_ID.cds";
			open(RNAFOLD,"$species_ID.$transcript_ID.RNAfold")||print LOG"Failed to open RNAfold output file $species_ID.$transcript_ID.RNAfold. $!\n";
			$head=<RNAFOLD>;
			$seq=<RNAFOLD>;
			$fold=<RNAFOLD>;
			close FOLD;
			unlink"$species_ID.$transcript_ID.RNAfold";
			unlink"$species_ID.${transcript_ID}_ss.ps";
			$fold=~s/\s*$//;
			$length=length$fold;
			$fold=~s/\.//g;
			$paired=length$fold;
			print OUT"orig_CDS\t$paired\n";
			}
		elsif($folding_tool==2) # use RNAplfold
			{
			system("RNAplfold < $species_ID.$transcript_ID.cds > RNAplfold_STDOUT.$species_ID.$transcript_ID");
			unlink"$species_ID.$transcript_ID.cds";
			unlink"RNAplfold_STDOUT.$species_ID.$transcript_ID";
			open(PS,"$species_ID.$transcript_ID.cds_dp.ps")||print LOG"Failed to open $species_ID.${transcript_ID}.cds_dp.ps. $!\n";
			%pairs=();
			while(<PS>)
				{
				$_=~s///;
				if($_=~/^\d/&&$_=~/ubox\s*$/)
					{
					@d=split(' ',$_);
					push(@{$pairs{$d[0]}},$d[2]);
					push(@{$pairs{$d[1]}},$d[2]);
					}
				}
			close PS;
			$paired=0;
			foreach$pos(keys%pairs)
				{
				@pp=@{$pairs{$pos}};
				$n_val=0;
				$sum=0;
				foreach$pp(@pp)
					{
					$n_val++;
					$sum+=$pp;
					}
				$paired+=($sum/$n_val);
				}
			$paired=int($paired+0.5);
			$paired=$paired*2; # pairing of one position means 2 paired bases
			
			# try hard to remove temporary files
			$t0=time;
			while(-e"$species_ID.$transcript_ID.cds_dp.ps")
				{
				unlink"$species_ID.$transcript_ID.cds_dp.ps";
				last if(time-$t0>2);
				}
			
			print OUT"orig_CDS\t$paired";
			if($keep_aORF){print OUT"\t$nucleotide_sequence";} # if optionally save alternative CDS sequence
			print OUT"\n";
			}
		if(!$silent){print"\n$ID2species{$species_ID}:$transcript_ID:origCDS -> $paired";}
		
		# start with construction of n artificial CDS sequences
		foreach$i(1..$alt_seqs)
			{
			$alt="";
			if($model==0) # FREE
				{
				foreach$pos(0..(length$protein_sequence)-1)
					{
					$n_codons=@{$a2c{substr($protein_sequence,$pos,1)}};
					$alt.=$a2c{substr($protein_sequence,$pos,1)}[int rand $n_codons];
					}
				}
			elsif($model==1) # CODON PROBABILITIES BASED ON ORIGINAL CODON USAGE
				{
				foreach$pos(0..(length$protein_sequence)-1)
					{
					$rand=rand;
					$sum_freq=0;
					foreach$possible_codon(@{$a2c{substr($protein_sequence,$pos,1)}})
						{
						$sum_freq+=$codon_frequency{$possible_codon};
						if($rand<=$sum_freq)
							{
							$alt.=$possible_codon;
							last;
							}
						}
					if($rand>$sum_freq) # for the unlikely case that sum of frequencies is <1 (due to rounding issues) and rand>sum of frequencies.
						{
						$n_codons=@{$a2c{substr($protein_sequence,$pos,1)}};
						$alt.=$a2c{substr($protein_sequence,$pos,1)}[int rand $n_codons];
						}
					}
				}
			elsif($model==2) # CODON SHUFFLE
				{
				%codon_dump_shuffle=();
				foreach$aa(keys%codon_dump)
					{
					$codon_dump{$aa}=~s/^#//;
					@codons=split('#',$codon_dump{$aa});
					@codons=shuffle@codons;
					foreach$codon(@codons)
						{
						$codon_dump_shuffle{$aa}.=$codon;
						}
					}
				foreach$pos(0..(length$protein_sequence)-1)
					{
					$codon_dump_shuffle{substr($protein_sequence,$pos,1)}=~s/^...//;
					$alt.=$&;
					}
				}
			open(ALT_CDS,">$species_ID.$transcript_ID.alt_cds${i}")||print LOG"Failed to create temporary input file $species_ID.$transcript_ID.alt_cds$i. $!\n";
			print ALT_CDS">$species_ID.$transcript_ID.alt_cds$i\n$alt";
			close ALT_CDS;
			
			# exit this fork if file opening failed
			if(!-e"$species_ID.$transcript_ID.alt_cds$i")
				{
				$pm->finish;
				}
			
			if($folding_tool==1) # use RNAfold
				{
				system("RNAfold < $species_ID.$transcript_ID.alt_cds${i} > $species_ID.$transcript_ID.alt_cds${i}.RNAfold");
				unlink"$species_ID.$transcript_ID.alt_cds${i}";
				open(RNAFOLD,"$species_ID.$transcript_ID.alt_cds${i}.RNAfold")||print LOG"Failed to open output file $species_ID.$transcript_ID.alt_cds${i}.RNAfold. $!\n";
				
				# exit this fork if RNAfold output file does not exist
				if(!-e"$species_ID.$transcript_ID.alt_cds${i}.RNAfold")
					{
					$pm->finish;
					}
				
				$head=<RNAFOLD>;
				$seq=<RNAFOLD>;
				$fold=<RNAFOLD>;
				close RNAFOLD;
				
				# try hard to remove temporary files
				$t0=time;
				while(-e"$species_ID.$transcript_ID.alt_cds${i}.RNAfold"||-e"$species_ID.$transcript_ID.alt_cds${i}_ss.ps")
					{
					unlink"$species_ID.$transcript_ID.alt_cds${i}.RNAfold";
					unlink"$species_ID.$transcript_ID.alt_cds${i}_ss.ps";
					last if(time-$t0>2);
					}
				
				$fold=~s/\s*$//;
				$length=length$fold;
				$fold=~s/\.//g;
				$paired=length$fold;
				}
			elsif($folding_tool==2)  # use RNAplfold
				{
				system("RNAplfold < $species_ID.$transcript_ID.alt_cds${i} > RNAplfold_STDOUT.$species_ID.$transcript_ID");
				unlink"$species_ID.$transcript_ID.alt_cds${i}";
				unlink"RNAplfold_STDOUT.$species_ID.$transcript_ID";
				open(PS,"$species_ID.$transcript_ID.alt_cds${i}_dp.ps")||print LOG"Failed to open output file $species_ID.$transcript_ID.alt_cds${i}_dp.ps. $!\n";
				
				# exit this fork if RNAplfold output file does not exist
				if(!-e"$species_ID.$transcript_ID.alt_cds${i}_dp.ps")
					{
					$pm->finish;
					}
				
				%pairs=();
				while(<PS>)
					{
					$_=~s///;
					if($_=~/^\d/&&$_=~/ubox\s*$/)
						{
						@d=split(' ',$_);
						push(@{$pairs{$d[0]}},$d[2]);
						push(@{$pairs{$d[1]}},$d[2]);
						}
					}
				close PS;
				$paired=0;
				foreach$pos(keys%pairs)
					{
					@pp=@{$pairs{$pos}};
					$n_val=0;
					$sum=0;
					foreach$pp(@pp)
						{
						$n_val++;
						$sum+=$pp;
						}
					$paired+=($sum/$n_val);
					}
				$paired=int($paired+0.5);
				$paired=$paired*2; # pairing at one position means 2 paired bases
				
				# try hard to remove temporary files
				$t0=time;
				while(-e"$species_ID.$transcript_ID.alt_cds${i}_dp.ps")
					{
					unlink"$species_ID.$transcript_ID.alt_cds${i}_dp.ps";
					last if(time-$t0>2);
					}
				}
			print OUT"altCDS_$i\t$paired";
			if($keep_aORF){print OUT"\t$alt";} # optionally save slternative CDS sequence
			print OUT"\n";
			
			if(!$silent){print"\n$ID2transcript{$species_ID}{$transcript_ID}:altCDS.${i} -> $paired";}
			}
		close OUT;
		$pm->finish;
		}
	$pm->wait_all_children;
	
	# remove remaining temporary files
	@remove_ps=glob("*.ps");
	@remove_RNAfold=glob("*.RNAfold");
	foreach$temp_file(@remove_ps,@remove_RNAfold)
		{
		unlink$temp_file;
		}
	}

# collect results
$LOG="$results_folder/PACKEIS.log";
$outfile_distr="$results_folder/DBFscores_distribution.txt";
$outfile_scores="$results_folder/DBFscores.txt";

%id2tr=();
open(LOG,$LOG)||die print$!;
while(<LOG>)
	{
	$_=~s/\s*$//;
	if($_=~s/^SPEC: //)
		{
		$_=~s/ .+$//;
		$spec=$_;
		print"\n$spec";
		$examples=0;
		}
	elsif($_=~s/^TR: //)
		{
		@d=split(" -> ",$_);
		$id2tr{$spec}{$d[1]}=$d[0];
		}
	}
close LOG;

%genes=();
%DBFscores=();
opendir(FOLDER,"$results_folder/raw")||print LOG"Failed to create output folder $results_folder/raw. $!.\n";
open(DISTR,">$outfile_distr")||print LOG"Failed to create output folder $results_folder/$outfile_distr. $!.\n";
open(SCORES,">$outfile_scores")||print LOG"Failed to create output folder $results_folder/$outfile_scores. $!.\n";
print SCORES"input_file\ttranscript\tDBF-score\n";

print DISTR"input_file";
$label=-0.01;
foreach(1..101)
	{
	$label+=0.01;
	$label=(int(($label*100)+0.5))/100; # on some systems there occur strange results like 0.800000000001
	print DISTR"\t$label";
	}
print DISTR"\n";
while($file=readdir FOLDER)
	{
	next if($file!~/\.results$/);
	$file=~s/\.\././;
	@d=split(/\./,$file);
	open(FILE,"$results_folder/raw/$file")||print LOG"Failed to open $results_folder/raw/$file. $!.\n";
	$head=<FILE>;
	@head=split("\t",$head);
	$more=0;
	$less=0;
	while(<FILE>)
		{
		$_=~s/\s*$//;
		@rep=split("\t",$_);
		if($rep[1]<$head[1])
			{
			$less++;
			}
		else
			{
			$more++;
			}
		}
	close FILE;
		
	if($less+$more<1)
		{
		print LOG"\n$results_folder/raw/$file seems to have no alt cds values!";
		next;
		}
	
	# Calculate DBF-score (0..1). High DBF-score means high backfolding compared to artificial CDS
	$DBFscore=(int(((($less/($less+$more))*100)+0.5)))/100;
	$DBFscores{$d[0]}{$DBFscore}++;
	
	# Save transcripts/genes sorted by DBF-score
	$genes{$d[0]}{$DBFscore}{$id2tr{$d[0]}{$d[1]}}=1;
	
	print SCORES"$d[0]\t$id2tr{$d[0]}{$d[1]}\t$DBFscore\n";
	}
close SCORES;

# print DBFscores to output file
foreach$organism(sort{$a cmp $b}keys%DBFscores)
	{
	print DISTR"$organism";
	foreach$i(0..100)
		{
		$DBFscore=$i/100;
		if(!$DBFscores{$organism}{$DBFscore})
			{
			$DBFscores{$organism}{$DBFscore}=0;
			}
		print DISTR"\t$DBFscores{$organism}{$DBFscore}";
		}
	print DISTR"\n";
	}
close DISTR;

# calculate codon usage per DBFscore for each species
foreach$spec(keys%genes)
	{
	%fasta=();
	open(FASTA,"$spec.fas")||die print$!;
	while(<FASTA>)
		{
		$_=~s/\s*$//;
		if($_=~s/^>//)
			{
			$tr=$_;
			}
		else
			{
			$_=~s/T/U/g;
			$fasta{$tr}.=uc$_;
			}
		}
	close FASTA;
	
	open(TABLE,">$results_folder/traits-per-DBFscore.$spec.table")||die print$!;
	print TABLE"DBFscore";
	foreach$aa(sort{$a cmp $b}keys%a2c)
		{
		foreach$codon(@{$a2c{$aa}})
			{
			print TABLE"\t$aa->$codon";
			}
		}
	print TABLE"\tavg.-GCcontent\tconcat.-GCcontent";
	
	foreach$aa(sort{$a cmp $b}keys%a2c)
		{
		print TABLE"\t$aa-fraction";
		}
	
	foreach$i(0..100)
		{
		$DBFscore=$i/100;
		print TABLE"\n$DBFscore";
		
		# Codon Usage and AA-composition
		%codon_counts=();
		%aa_counts=();
		foreach$tr(keys%{$genes{$spec}{$DBFscore}})
			{
			$codon_start=-3;
			while(1)
				{
				$codon_start+=3;
				$codon_counts{substr($fasta{$tr},$codon_start,3)}++;
				$aa_counts{$c2a{substr($fasta{$tr},$codon_start,3)}}++;
				last if(!$c2a{substr($fasta{$tr},$codon_start,3)}); # stop with stop codon
				last if($codon_start>=length$fasta{$tr}); # stop at end of transcript
				}
			}
		%codon_frequency=();
		foreach$aa(sort{$a cmp $b}keys%a2c)
			{
			foreach$codon(@{$a2c{$aa}})
				{
				$freq="n.a.";
				if($aa_counts{$aa}>0)
					{
					$codon_frequency{$codon}=$codon_counts{$codon}/$aa_counts{$aa};
					$freq=$codon_counts{$codon}/$aa_counts{$aa};
					print TABLE"\t$freq";
					}
				}
			}
		
		# GC content
		$c=0;
		$avgGCcontent=0;
		$concat="";
		foreach$tr(keys%{$genes{$spec}{$DBFscore}})
			{
			$c++;
			$concat.=$fasta{$tr};
			$length=length$fasta{$tr};
			$fasta{$tr}=~s/[GC]//g;
			$avgGCcontent+=1-((length$fasta{$tr})/$length);
			}
		
		if($c>0) # Possibly there exist no genes with the current DBF-score
			{
			$avgGCcontent=$avgGCcontent/$c;
			$length=length$concat;
			$concat=~s/[GC]//g;
			$GCconcat=1-((length$concat)/$length);
			print TABLE"\t$avgGCcontent\t$GCconcat";
			}
		else
			{
			print TABLE"\tn.a.\tn.a.";
			}
		
		# print AA-composition
		$total_aa=0;
		foreach$aa(sort{$a cmp $b}keys%a2c)
			{
			$total_aa+=$aa_counts{$aa};
			}
		if($total_aa>0)
			{
			foreach$aa(sort{$a cmp $b}keys%a2c)
				{
				$aa_frac=$aa_counts{$aa}/$total_aa;
				print TABLE"\t$aa_frac";
				}
			}
		
		undef%codon_counts;
		undef%aa_counts;
		undef%codon_frequency;
		}
	undef%fasta;
	close TABLE;
	}
undef%id2tr;
undef%DBFscores;
undef%genes;


# STANDARD GENETIC CODE
sub a2c # Aminoacid -> Codon(s)
	{
	%a2c=
		(
		'F'=>['UUU','UUC'],
		'L'=>['UUA','UUG','CUU','CUC','CUA','CUG'],
		'I'=>['AUU','AUC','AUA'],
		'M'=>['AUG'],
		'V'=>['GUU','GUC','GUA','GUG'],
		'S'=>['UCU','UCC','UCA','UCG','AGU','AGC'],
		'P'=>['CCU','CCC','CCA','CCG'],
		'T'=>['ACU','ACC','ACA','ACG'],
		'A'=>['GCU','GCC','GCA','GCG'],
		'Y'=>['UAU','UAC'],
		'H'=>['CAU','CAC'],
		'Q'=>['CAA','CAG'],
		'N'=>['AAU','AAC'],
		'K'=>['AAA','AAG'],
		'D'=>['GAU','GAC'],
		'E'=>['GAA','GAG'],
		'C'=>['UGU','UGC'],
		'W'=>['UGG'],
		'R'=>['CGU','CGC','CGA','CGG','AGA','AGG'],
		'G'=>['GGU','GGC','GGA','GGG'],
		);
	}

sub c2a # Codon -> Aminoacid
	{
	%c2a=
		(
		'UUU'=>'F','UUC'=>'F',
		'UUA'=>'L','UUG'=>'L','CUU'=>'L','CUC'=>'L','CUA'=>'L','CUG'=>'L',
		'AUU'=>'I','AUC'=>'I','AUA'=>'I',
		'AUG'=>'M',
		'GUU'=>'V','GUA'=>'V','GUC'=>'V','GUG'=>'V',
		'UCU'=>'S','UCC'=>'S','UCA'=>'S','UCG'=>'S','AGU'=>'S','AGC'=>'S',
		'CCU'=>'P','CCC'=>'P','CCA'=>'P','CCG'=>'P',
		'ACU'=>'T','ACC'=>'T','ACA'=>'T','ACG'=>'T',
		'GCU'=>'A','GCC'=>'A','GCA'=>'A','GCG'=>'A',
		'UAU'=>'Y','UAC'=>'Y',
		'CAU'=>'H','CAC'=>'H',
		'CAA'=>'Q','CAG'=>'Q',
		'AAU'=>'N','AAC'=>'N',
		'AAA'=>'K','AAG'=>'K',
		'GAU'=>'D','GAC'=>'D',
		'GAA'=>'E','GAG'=>'E',
		'UGU'=>'C','UGC'=>'C',
		'UGG'=>'W',
		'CGU'=>'R','CGC'=>'R','CGA'=>'R','CGG'=>'R','AGA'=>'R','AGG'=>'R',
		'GGU'=>'G','GGC'=>'G','GGA'=>'G','GGG'=>'G',
		);
	}

# MITOCHONDRIAL GENETIC CODE
sub a2c_MT # Aminoacid -> Codon(s)
	{
	%a2c=
		(
		'F'=>['UUU','UUC'],
		'L'=>['UUA','UUG','CUU','CUC','CUA','CUG'],
		'I'=>['AUU','AUC'],
		'M'=>['AUG','AUA'],
		'V'=>['GUU','GUC','GUA','GUG'],
		'S'=>['UCU','UCC','UCA','UCG','AGU','AGC'],
		'P'=>['CCU','CCC','CCA','CCG'],
		'T'=>['ACU','ACC','ACA','ACG'],
		'A'=>['GCU','GCC','GCA','GCG'],
		'Y'=>['UAU','UAC'],
		'H'=>['CAU','CAC'],
		'Q'=>['CAA','CAG'],
		'N'=>['AAU','AAC'],
		'K'=>['AAA','AAG'],
		'D'=>['GAU','GAC'],
		'E'=>['GAA','GAG'],
		'C'=>['UGU','UGC'],
		'W'=>['UGG','UGA'],
		'R'=>['CGU','CGC','CGA','CGG'],
		'G'=>['GGU','GGC','GGA','GGG'],
		);
	}

sub c2a_MT # Codon -> Aminoacid
	{
	%c2a=
		(
		'UUU'=>'F','UUC'=>'F',
		'UUA'=>'L','UUG'=>'L','CUU'=>'L','CUC'=>'L','CUA'=>'L','CUG'=>'L',
		'AUU'=>'I','AUC'=>'I',
		'AUG'=>'M','AUA'=>'M',
		'GUU'=>'V','GUA'=>'V','GUC'=>'V','GUG'=>'V',
		'UCU'=>'S','UCC'=>'S','UCA'=>'S','UCG'=>'S','AGU'=>'S','AGC'=>'S',
		'CCU'=>'P','CCC'=>'P','CCA'=>'P','CCG'=>'P',
		'ACU'=>'T','ACC'=>'T','ACA'=>'T','ACG'=>'T',
		'GCU'=>'A','GCC'=>'A','GCA'=>'A','GCG'=>'A',
		'UAU'=>'Y','UAC'=>'Y',
		'CAU'=>'H','CAC'=>'H',
		'CAA'=>'Q','CAG'=>'Q',
		'AAU'=>'N','AAC'=>'N',
		'AAA'=>'K','AAG'=>'K',
		'GAU'=>'D','GAC'=>'D',
		'GAA'=>'E','GAG'=>'E',
		'UGU'=>'C','UGC'=>'C',
		'UGG'=>'W','UGA'=>'W',
		'CGU'=>'R','CGC'=>'R','CGA'=>'R','CGG'=>'R',
		'GGU'=>'G','GGC'=>'G','GGA'=>'G','GGG'=>'G',
		);
	}
